# -*- coding: utf-8 -*-
# Copyright (C) 2017-2018 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
from IPy import IP

from ..error import ConfigError
from ..setting import undefined, OptionBag, Undefined
from ..i18n import _
from .option import Option


class NetmaskOption(Option):
    "represents the choice of a netmask"
    __slots__ = tuple()
    _display_name = _('netmask address')

    def _validate(self,
                  value: str,
                  option_bag: OptionBag,
                  current_opt: Option=Undefined) -> None:
        if not isinstance(value, str):
            raise ValueError(_('invalid string'))
        if value.count('.') != 3:
            raise ValueError()
        for val in value.split('.'):
            if val.startswith("0") and len(val) > 1:
                raise ValueError()
        try:
            IP('0.0.0.0/{0}'.format(value))
        except ValueError:
            raise ValueError()

    def _cons_network_netmask(self,
                              current_opt,
                              opts,
                              vals,
                              warnings_only,
                              context):
        #opts must be (netmask, network) options
        if context is undefined and len(vals) != 2:
            raise ConfigError(_('network_netmask needs a network and a netmask'))
        if None in vals or len(vals) != 2:
            return
        return self.__cons_netmask(current_opt,
                                   opts,
                                   vals[0],
                                   vals[1],
                                   False,
                                   warnings_only,
                                   'network')

    def _cons_ip_netmask(self,
                         current_opt,
                         opts,
                         vals,
                         warnings_only,
                         context):
        #opts must be (netmask, ip) options
        if context is undefined and len(vals) != 2:
            raise ConfigError(_('ip_netmask needs an IP and a netmask'))
        if None in vals or len(vals) != 2:
            return
        self.__cons_netmask(current_opt,
                            opts,
                            vals[0],
                            vals[1],
                            True,
                            warnings_only,
                            'ip')

    def __cons_netmask(self,
                       current_opt,
                       opts,
                       val_netmask,
                       val_ipnetwork,
                       make_net,
                       warnings_only,
                       typ):
        msg = None
        try:
            ip = IP('{0}/{1}'.format(val_ipnetwork, val_netmask), make_net=make_net)
            #if cidr == 32, ip same has network
            if make_net and ip.prefixlen() != 32:
                val_ip = IP(val_ipnetwork)
                if ip.net() == val_ip:
                    if current_opt == opts[1]:
                        msg = _('this is a network with netmask "{0}" ("{1}")')
                    else:
                        msg = _('this is a network with {2} "{0}" ("{1}")')
                elif ip.broadcast() == val_ip:
                    if current_opt == opts[1]:
                        msg = _('this is a broadcast with netmask "{0}" ("{1}")')
                    else:
                        msg = _('this is a broadcast with {2} "{0}" ("{1}")')

        except ValueError:
            if not make_net:
                if current_opt == opts[1]:
                    raise ValueError(_('with netmask "{0}" ("{1}")').format(val_netmask, opts[0].impl_get_display_name()))
                else:
                    raise ValueError(_('with {2} "{0}" ("{1}")').format(val_ipnetwork, opts[1].impl_get_display_name(), typ))
        if msg is not None:
            if current_opt == opts[1]:
                raise ValueError(msg.format(val_netmask,
                                            opts[1].impl_get_display_name()))
            else:
                raise ValueError(msg.format(val_ipnetwork,
                                            opts[0].impl_get_display_name(),
                                            typ))
